/*
 * Copyright (c) Huawei Technologies Co., Ltd. 2020-2024. All rights reserved.
 */

#ifndef JAC_AP_FUSIONSERVICE_FUSIONSERVICE_SKELETON_H
#define JAC_AP_FUSIONSERVICE_FUSIONSERVICE_SKELETON_H

#include "ara/com/internal/skeleton/skeleton_adapter.h"
#include "ara/com/internal/skeleton/event_adapter.h"
#include "ara/com/internal/skeleton/field_adapter.h"
#include "ara/com/internal/skeleton/method_adapter.h"
#include "ara/com/crc_verification.h"
#include "jac/ap/fusionservice/fusionservice_common.h"
#include <cstdint>

namespace jac {
namespace ap {
namespace fusionservice {
namespace skeleton {
namespace events
{
    using ObjectEvent = ara::com::internal::skeleton::event::EventAdapter<::dpc::TrafficParticipantStruct>;
    using Fusion2VizRulebasedResultEvent = ara::com::internal::skeleton::event::EventAdapter<::PerceptionBoundingBoxArray>;
    static constexpr ara::com::internal::EntityId FusionServiceObjectEventId = 35544U; //ObjectEvent_event_hash
    static constexpr ara::com::internal::EntityId FusionServiceFusion2VizRulebasedResultEventId = 55454U; //Fusion2VizRulebasedResultEvent_event_hash
}

namespace methods
{
}

namespace fields
{
}

class FusionServiceSkeleton {
private:
    std::unique_ptr<ara::com::internal::skeleton::SkeletonAdapter> skeletonAdapter;
    void ConstructSkeleton(const ara::com::MethodCallProcessingMode mode)
    {
        if (mode == ara::com::MethodCallProcessingMode::kEvent) {
            if (!(skeletonAdapter->SetMethodThreadNumber(skeletonAdapter->GetMethodThreadNumber(0U), 1024U))) {
#ifndef NOT_SUPPORT_EXCEPTIONS
                ara::core::ErrorCode errorcode(ara::com::ComErrc::kNetworkBindingFailure);
                throw ara::com::ComException(std::move(errorcode));
#else
                std::cerr << "Error: Not support exception, create skeleton failed!" << std::endl;
#endif
            }
        }
        const ara::core::Result<void> resultObjectEvent = ObjectEvent.Initialize();
        ThrowError(resultObjectEvent);
        const ara::core::Result<void> resultFusion2VizRulebasedResultEvent = Fusion2VizRulebasedResultEvent.Initialize();
        ThrowError(resultFusion2VizRulebasedResultEvent);
    }

    FusionServiceSkeleton& operator=(const FusionServiceSkeleton&) = delete;

    static void ThrowError(const ara::core::Result<void>& result)
    {
        if (!(result.HasValue())) {
#ifndef NOT_SUPPORT_EXCEPTIONS
            ara::core::ErrorCode errorcode(result.Error());
            throw ara::com::ComException(std::move(errorcode));
#else
            std::cerr << "Error: Not support exception, create skeleton failed!" << std::endl;
#endif
        }
    }
public:
    class ConstructionToken {
    public:
        explicit ConstructionToken(const ara::com::InstanceIdentifier& instanceId,
            const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
            : processMode(mode),
              ptr(std::make_unique<ara::com::internal::skeleton::SkeletonAdapter>(::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceId, mode)),
              ObjectEvent(ptr->GetSkeleton(), events::FusionServiceObjectEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceId),
              Fusion2VizRulebasedResultEvent(ptr->GetSkeleton(), events::FusionServiceFusion2VizRulebasedResultEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceId){
        }

        explicit ConstructionToken(const ara::core::InstanceSpecifier& instanceSpec,
            const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
            : processMode(mode),
              ptr(std::make_unique<ara::com::internal::skeleton::SkeletonAdapter>(::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceSpec, mode)),
              ObjectEvent(ptr->GetSkeleton(), events::FusionServiceObjectEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceSpec),
              Fusion2VizRulebasedResultEvent(ptr->GetSkeleton(), events::FusionServiceFusion2VizRulebasedResultEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceSpec){
        }

        explicit ConstructionToken(const ara::com::InstanceIdentifierContainer instanceContainer,
            const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
            : processMode(mode),
              ptr(std::make_unique<ara::com::internal::skeleton::SkeletonAdapter>(::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceContainer, mode)),
              ObjectEvent(ptr->GetSkeleton(), events::FusionServiceObjectEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceContainer),
              Fusion2VizRulebasedResultEvent(ptr->GetSkeleton(), events::FusionServiceFusion2VizRulebasedResultEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceContainer){
        }

        ConstructionToken(ConstructionToken&& other)
            : processMode(std::move(other.processMode)),
              ptr(std::move(other.ptr)),
              ObjectEvent(std::move(other.ObjectEvent)),
              Fusion2VizRulebasedResultEvent(std::move(other.Fusion2VizRulebasedResultEvent)){
        }
        ConstructionToken& operator=(ConstructionToken && other)
        {
            if (&other != this) {
                processMode = std::move(other.processMode);
                ptr = std::move(other.ptr);
                ObjectEvent = std::move(other.ObjectEvent);
                Fusion2VizRulebasedResultEvent = std::move(other.Fusion2VizRulebasedResultEvent);
            }
            return *this;
        }
        ConstructionToken(const ConstructionToken&) = delete;
        ConstructionToken& operator = (const ConstructionToken&) = delete;
        std::unique_ptr<ara::com::internal::skeleton::SkeletonAdapter> GetSkeletonAdapter()
        {
            return std::move(ptr);
        }
        events::ObjectEvent GetObjectEvent()
        {
            return std::move(ObjectEvent);
        }
        events::Fusion2VizRulebasedResultEvent GetFusion2VizRulebasedResultEvent()
        {
            return std::move(Fusion2VizRulebasedResultEvent);
        }
        ara::core::Result<void> Initialize()
        {
            ara::core::Result<void> initResult;
            do {
                if (processMode == ara::com::MethodCallProcessingMode::kEvent) {
                    if(!ptr->SetMethodThreadNumber(ptr->GetMethodThreadNumber(0U), 1024U)) {
                        ara::core::ErrorCode errorCode(ara::com::ComErrc::kNetworkBindingFailure);
                        initResult.EmplaceError(errorCode);
                        break;
                    }
                }
                initResult = ObjectEvent.Initialize();
                if (!initResult.HasValue()) {
                    break;
                }
                initResult = Fusion2VizRulebasedResultEvent.Initialize();
                if (!initResult.HasValue()) {
                    break;
                }
            } while(false);

            return initResult;
        }
    private:
        ara::com::MethodCallProcessingMode processMode;
        std::unique_ptr<ara::com::internal::skeleton::SkeletonAdapter> ptr;
        events::ObjectEvent ObjectEvent;
        events::Fusion2VizRulebasedResultEvent Fusion2VizRulebasedResultEvent;
    };
    explicit FusionServiceSkeleton(const ara::com::InstanceIdentifier& instanceId,
                           const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
        : skeletonAdapter(std::make_unique<ara::com::internal::skeleton::SkeletonAdapter>(::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceId, mode)),
          ObjectEvent(skeletonAdapter->GetSkeleton(), events::FusionServiceObjectEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceId),
          Fusion2VizRulebasedResultEvent(skeletonAdapter->GetSkeleton(), events::FusionServiceFusion2VizRulebasedResultEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceId){
        ConstructSkeleton(mode);
    }

    explicit FusionServiceSkeleton(const ara::core::InstanceSpecifier& instanceSpec,
                           const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
        :skeletonAdapter(std::make_unique<ara::com::internal::skeleton::SkeletonAdapter>(::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceSpec, mode)),
          ObjectEvent(skeletonAdapter->GetSkeleton(), events::FusionServiceObjectEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceSpec),
          Fusion2VizRulebasedResultEvent(skeletonAdapter->GetSkeleton(), events::FusionServiceFusion2VizRulebasedResultEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceSpec){
        ConstructSkeleton(mode);
    }

    explicit FusionServiceSkeleton(const ara::com::InstanceIdentifierContainer instanceContainer,
                           const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
        :skeletonAdapter(std::make_unique<ara::com::internal::skeleton::SkeletonAdapter>(::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceContainer, mode)),
          ObjectEvent(skeletonAdapter->GetSkeleton(), events::FusionServiceObjectEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceContainer),
          Fusion2VizRulebasedResultEvent(skeletonAdapter->GetSkeleton(), events::FusionServiceFusion2VizRulebasedResultEventId, ::jac::ap::fusionservice::FusionService::ServiceIdentifier, instanceContainer){
        ConstructSkeleton(mode);
    }

    FusionServiceSkeleton(const FusionServiceSkeleton&) = delete;

    FusionServiceSkeleton(FusionServiceSkeleton&&) = default;
    FusionServiceSkeleton& operator=(FusionServiceSkeleton&&) = default;
    FusionServiceSkeleton(ConstructionToken&& token) noexcept
        : skeletonAdapter(token.GetSkeletonAdapter()),
          ObjectEvent(token.GetObjectEvent()),
          Fusion2VizRulebasedResultEvent(token.GetFusion2VizRulebasedResultEvent()){
    }

    static ara::core::Result<ConstructionToken> Preconstruct(
        ara::com::InstanceIdentifier instanceId,
        const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
    {
        ConstructionToken token(instanceId, mode);
        const auto initResult = token.Initialize();
        ara::core::Result<ConstructionToken> result(std::move(token));
        if (!initResult.HasValue()) {
            const ara::core::ErrorCode errorCode(initResult.Error());
            result.EmplaceError(errorCode);
        }
        return result;
    }

    static ara::core::Result<ConstructionToken> Preconstruct(
        ara::core::InstanceSpecifier instanceSpec,
        const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
    {
        ConstructionToken token(instanceSpec, mode);
        const auto initResult = token.Initialize();
        ara::core::Result<ConstructionToken> result(std::move(token));
        if (!initResult.HasValue()) {
            const ara::core::ErrorCode errorCode(initResult.Error());
            result.EmplaceError(errorCode);
        }
        return result;
    }

    static ara::core::Result<ConstructionToken> Preconstruct(
        ara::com::InstanceIdentifierContainer instanceIdContainer,
        const ara::com::MethodCallProcessingMode mode = ara::com::MethodCallProcessingMode::kEvent)
    {
        ConstructionToken token(instanceIdContainer, mode);
        const auto initResult = token.Initialize();
        ara::core::Result<ConstructionToken> result(std::move(token));
        if (!initResult.HasValue()) {
            const ara::core::ErrorCode errorCode(initResult.Error());
            result.EmplaceError(errorCode);
        }
        return result;
    }

    virtual ~FusionServiceSkeleton()
    {
        StopOfferService();
    }

    void OfferService()
    {
        skeletonAdapter->RegisterE2EErrorHandler(&FusionServiceSkeleton::E2EErrorHandler, *this);
        skeletonAdapter->OfferService();
    }
    void StopOfferService()
    {
        skeletonAdapter->StopOfferService();
    }
    ara::core::Future<bool> ProcessNextMethodCall()
    {
        return skeletonAdapter->ProcessNextMethodCall();
    }
    bool SetMethodThreadNumber(const std::uint16_t& number, const std::uint16_t& queueSize)
    {
        return skeletonAdapter->SetMethodThreadNumber(number, queueSize);
    }

    virtual void E2EErrorHandler(ara::com::e2e::E2EErrorCode, ara::com::e2e::DataID, ara::com::e2e::MessageCounter){}

    events::ObjectEvent ObjectEvent;
    events::Fusion2VizRulebasedResultEvent Fusion2VizRulebasedResultEvent;
};
} // namespace skeleton
} // namespace fusionservice
} // namespace ap
} // namespace jac

#endif // JAC_AP_FUSIONSERVICE_FUSIONSERVICE_SKELETON_H
